package main

import (
	"bytes"
	"encoding/hex"
	"os"
	"path/filepath"
	"strings"
	"testing"
)

func TestPackageErrors_Add(t *testing.T) {
	eh := &PackageErrors{}
	p := mockPkg()

	eh.Add(p, "some reason")
	if len(*eh) != 1 {
		t.Errorf("expected 1 error message, got %d", len(*eh))
	}

	want := "name: some reason"
	if (*eh)[0] != want {
		t.Errorf("expected error '%s', got '%s'", want, (*eh)[0])
	}
}

func TestPackageErrors_SkipPkgAndDependees(t *testing.T) {
	eh := &PackageErrors{}
	p := mockPkg() // name = "name", dependees = [dependee1, dependee2]

	eh.SkipPkgAndDependees(p, "skipped reason")

	if len(*eh) != 3 {
		t.Errorf("expected 3 error messages, got %d", len(*eh))
	}
	// The order should be:
	//   1) name: skipped reason
	//   2) dependee1: Depends on name
	//   3) dependee2: Depends on name

	want1 := "name: skipped reason"
	want2 := "prev: Depends on name"
	want3 := "next: Depends on name"

	if (*eh)[0] != want1 {
		t.Errorf("expected '%s', got '%s'", want1, (*eh)[0])
	}
	if (*eh)[1] != want2 {
		t.Errorf("expected '%s', got '%s'", want2, (*eh)[1])
	}
	if (*eh)[2] != want3 {
		t.Errorf("expected '%s', got '%s'", want3, (*eh)[2])
	}

	// Ensure p is removed from the list:
	if p.IsInList() {
		t.Errorf("expected package to be removed from the list, but it's still in list")
	}
}

func TestPackageErrors_CheckDownloads_FileNotFound(t *testing.T) {
	eh := &PackageErrors{}

	head := &Package{}
	p := mockPkg()
	p.Append(head)

	createPkgTestfiles(p, t)

	eh.CheckDownloads(head, p.path)

	if len(*eh) != 3 { // 1 for the package, 2 for the dependees
		t.Fatalf("expected 1 error message, got %d. Errors: \n%s", len(*eh), *eh)
	}

	want := filepath.Join(p.path, "foo.tar.gz") + ": file not found"
	if !strings.Contains((*eh)[0], want) {
		t.Errorf("expected error to contain '%s', got '%s'", want, (*eh)[0])
	}
}

func TestPackageErrors_CheckDownloads_Success(t *testing.T) {
	eh := &PackageErrors{}

	head := &Package{}
	p := mockPkg() // e.g., returns a *Package with p.name set, etc.
	p.prev, p.next = nil, nil
	p.Append(head)

	// We want to create a dummy file that matches the expected MD5.
	// Let's pick some known file content:
	fileContent := []byte("Hello from the test file")
	correctMD5 := hex.EncodeToString(md5Sum(bytes.NewReader(fileContent)))

	// Suppose the package is expecting to download a file named "foo.tar.gz".
	downloadName := "foo.tar.gz"

	// Set up the .info file with the correct MD5.
	createTestfileArgs(
		p,
		t,
		downloadName, // DOWNLOAD
		correctMD5,   // MD5SUM
		"",           // DOWNLOAD_x86_64
		"",           // MD5SUM_x86_64
		"",           // REQUIRES
	)

	// Now actually create the "download" file in p.path with the matching content.
	fullPath := filepath.Join(p.path, downloadName)
	if err := os.WriteFile(fullPath, fileContent, 0644); err != nil {
		t.Fatalf("failed to create test file: %v", err)
	}

	// Run the check
	eh.CheckDownloads(head, p.path)

	// If everything is correct, we expect zero errors
	if len(*eh) != 0 {
		t.Fatalf("expected no errors, got %d errors. Errors:\n%s", len(*eh), *eh)
	}
}

func TestPackageErrors_CheckDownloads_MD5Mismatch(t *testing.T) {
	eh := &PackageErrors{}

	head := &Package{}
	p := mockPkg()
	p.Append(head)

	// Let’s pick some content that won't match the expected MD5
	fileContent := []byte("Hello from mismatch land")

	// We'll purposely supply a WRONG MD5 sum to the .info file
	wrongMD5 := "00000000000000000000000000000000" // definitely not correctMD5

	downloadName := "foo.tar.gz"

	// Write the .info file with the WRONG MD5
	createTestfileArgs(
		p,
		t,
		downloadName,
		wrongMD5,
		"",
		"",
		"",
	)

	// Actually create the file with the "correctMD5" content (which doesn't match wrongMD5)
	fullPath := filepath.Join(p.path, downloadName)
	if err := os.WriteFile(fullPath, fileContent, 0644); err != nil {
		t.Fatalf("failed to create test file: %v", err)
	}

	// Now this should cause a mismatch error
	eh.CheckDownloads(head, p.path)

	// 1 error for the package, plus errors for each "dependee" => total 3.
	if len(*eh) != 3 {
		t.Fatalf("expected 3 errors (1 for the package + 2 for dependees), got %d. Errors:\n%s", len(*eh), *eh)
	}

	// Check the first error message mentions 'MD5Sum Mismatch'
	want := filepath.Join(p.path, downloadName) + ": MD5Sum Mismatch"
	if !strings.Contains((*eh)[0], want) {
		t.Errorf("expected error to contain '%s', got '%s'", want, (*eh)[0])
	}
}

func TestPackageErrors_Final_NoErrors(t *testing.T) {
	eh := &PackageErrors{}
	var buf bytes.Buffer

	eh.Final(&buf)
	got := buf.String()
	want := "All packages installed successfully\n"
	if got != want {
		t.Errorf("expected '%s', got '%s'", want, got)
	}
}

func TestPackageErrors_Final_WithErrors(t *testing.T) {
	eh := &PackageErrors{
		"some error 1",
		"some error 2",
	}
	var buf bytes.Buffer

	eh.Final(&buf)
	got := buf.String()

	// We expect the Banner(...) output
	if !strings.Contains(got, "Skipped Packages") {
		t.Error("expected output to contain 'Skipped Packages'")
	}
	if !strings.Contains(got, "some error 1") {
		t.Error("expected output to contain 'some error 1'")
	}
	if !strings.Contains(got, "some error 2") {
		t.Error("expected output to contain 'some error 2'")
	}
}
