package main

import (
	"bufio"
	"bytes"
	"crypto/md5"
	"encoding/hex"
	"os"
	"path/filepath"
	"regexp"
	"strings"
	"testing"
)

func TestGetKernelVersion(t *testing.T) {
	VersionPattern := regexp.MustCompile(`\d+\.\d+\.\d+`)
	// Test the GetKernelVersion function
	if GetKernelVersion() == "" {
		t.Error("GetKernelVersion() returned an empty string")
	}
	if !VersionPattern.MatchString(GetKernelVersion()) {
		t.Error("GetKernelVersion() returned an invalid version")
	}
	t.Log("GetKernelVersion() test passed with version:", GetKernelVersion())
}

func TestStripCommentsAndTokenize(t *testing.T) {
	input := "foo bar # baz\nqux"
	tokens := StripCommentsAndTokenize(strings.NewReader(input))
	if len(tokens) != 3 {
		t.Error("StripCommentsAndTokenize() returned an incorrect number of tokens")
	}
	if tokens[0] != "foo" || tokens[1] != "bar" || tokens[2] != "qux" {
		t.Error("StripCommentsAndTokenize() returned incorrect tokens")
	}
	t.Log("StripCommentsAndTokenize() test passed with tokens:", tokens)
}

func TestIsMD5SumMatch(t *testing.T) {
	t.Run("Matching MD5", func(t *testing.T) {
		// Create a temporary file with known content
		tempDir := t.TempDir()
		filePath := filepath.Join(tempDir, "testfile.txt")
		content := []byte("Hello, world!\n")
		if err := os.WriteFile(filePath, content, 0o644); err != nil {
			t.Fatalf("failed to write testfile.txt: %v", err)
		}

		// Compute MD5 of that content ourselves
		sum := md5.Sum(content)
		md5Str := hex.EncodeToString(sum[:])

		// Now call IsMD5SumMatch with the correct MD5
		match := IsMD5SumMatch(md5Str, filePath)
		if !match {
			t.Errorf("expected true for correct MD5, got false")
		}
	})

	t.Run("Non-Matching MD5", func(t *testing.T) {
		tempDir := t.TempDir()
		filePath := filepath.Join(tempDir, "testfile.txt")
		content := []byte("FooBar\n")
		if err := os.WriteFile(filePath, content, 0o644); err != nil {
			t.Fatalf("failed to write testfile.txt: %v", err)
		}

		// Compute MD5 of something else
		dummySum := md5.Sum([]byte("Some other content"))
		md5Str := hex.EncodeToString(dummySum[:])

		match := IsMD5SumMatch(md5Str, filePath)
		if match {
			t.Errorf("expected false for incorrect MD5, got true")
		}
	})

	t.Run("File Does Not Exist", func(t *testing.T) {
		// Depending on how your `open(filename)` handles a missing file,
		// this may panic or return false.
		// Example if it panics, we can capture with defer recover().
		defer func() {
			if r := recover(); r == nil {
				t.Error("expected a panic when file does not exist, got no panic")
			}
		}()

		fakePath := filepath.Join(t.TempDir(), "missing.txt")
		_ = IsMD5SumMatch("dummyMD5", fakePath)
		t.Error("should have panicked before this line (file not found)")
	})
}

func TestPkgNamesAndBuildOptions(t *testing.T) {
	input := []string{"foo", "bar=baz"}
	names, buildOptions := PkgNamesAndBuildOptions(input)
	if len(names) != 1 || names[0] != "foo" {
		t.Error("PkgNamesAndBuildOptions() returned incorrect package names")
	}
	if len(buildOptions) != 1 || buildOptions["foo"][0] != "bar=baz" {
		t.Error("PkgNamesAndBuildOptions() returned incorrect build options")
	}
	t.Log("PkgNamesAndBuildOptions() test passed with names:", names, "and build options:", buildOptions)
}

func TestPromptUser(t *testing.T) {
	in := strings.NewReader("foo\n")
	var out bytes.Buffer
	buf := bufio.NewWriter(&out)
	prompt := "bar"
	res := PromptUser(buf, in, prompt)
	buf.Flush()
	if out.String() != prompt+" [y/N]: " {
		t.Error("PromptUser() did not write the correct prompt")
	}
	if res {
		t.Error("PromptUser() returned true for 'foo'")
	}
	if !PromptUser(buf, strings.NewReader("y\n"), prompt) {
		t.Error("PromptUser() returned false for 'y'")
	}
	if !PromptUser(buf, strings.NewReader("Y\n"), prompt) {
		t.Error("PromptUser() returned false for 'Y'")
	}
	r := strings.NewReader("N\nT")
	PromptUser(buf, r, prompt)
	if ReadChar(r) != 'T' {
		t.Error("Newline not correctly consumed")
	}
}

func TestJoin(t *testing.T) {
	got, want := Join([]string{"foo", "bar"}), "foo bar"
	if got != want {
		t.Error("Join() returned an incorrect path")
	}
	t.Log("Join() test passed with path: ", got)
}
