package main

import (
	"bufio"
	"log"
	"os"
	"path/filepath"
	"regexp"
	"strings"
)

func main() {
	var ( // Set up everything that does io
		stdout   = bufio.NewWriter(os.Stdout)
		logger   = log.Default()
		executor = Executor{os.Stdout, os.Stderr, logger} // To run external commands
		pager    = NewPager(&executor)                    // To display files
		settings = NewSettings(os.Args, DEFAULT_PKG_FILE) // Parse command line arguments
	)
	defer stdout.Flush()
	logger.SetOutput(stdout)

	if !Exists(REPO_PATH) { // Make sure the repository exists
		Mkdir(filepath.Dir(REPO_PATH))
		executor.PanicOnFail(PULL_CMD)
	}

	if settings.Update() {
		executor.PanicOnFail(SYNC_CMD)
	}

	pMap := NewRepoMap(REPO_PATH) // Read all packages from the repository

	switch settings.Cmd() {

	case VIEW:
		for _, pName := range settings.PkgNames() {
			pkg := pMap.Get(pName)
			pager.Display(pkg.AllReadmes() + " " + pkg.InfoFile())
		}

	case VIEW_BUILD:
		for _, pName := range settings.PkgNames() {
			pager.Display(pMap.Get(pName).SlackBuild())
		}

	case SEARCH:
		for _, searchString := range settings.PkgNames() {
			re := regexp.MustCompile(searchString)
			for pName := range pMap {
				if re.MatchString(pName) {
					stdout.WriteString(pMap.Get(pName).Description() + "\n")
				}
			}
		}

	case DEPENDEES:
		pMap.ProcessList(settings.PkgNames())
		pMap.Get(settings.Root()).IterateDependees(
			func(pkg *Package) { stdout.WriteString(pkg.Name() + "\n") })

	case TRACKING:
		pMap.ProcessList(settings.PkgNames()).Iterate(
			func(pkg *Package) { stdout.WriteString(pkg.Name() + "\n") })

	case LIST_INSTALLED:
		for _, file := range ReadDir(DOC_DIR) {
			pName := strings.Split(file.Name(), "-")[0]
			if pMap.Contains(pName) {
				stdout.WriteString(pName + "\n")
			}
		}

	case UPGRADE:
		var (
			pErr  PackageErrors
			pList Head
		)

		if settings.NoDeps() { // Upgrade only the packages in the list
			pList = pMap.PkgNames2List(settings.PkgNames())
		} else { // Upgrade the packages in the list and their dependencies
			pList = pMap.ProcessList(settings.PkgNames())
		}

		// Call after creation of pList since if this panics the handler would write
		// a misleading success message
		defer pErr.Final(stdout)

		if settings.UpdateBlacklist() {
			if err := pList.UpdateBlacklist(
				BLACKLIST,
				OTHER_BLACKLIST,
				DEFAULT_PKG_FILE,
			); err != nil {
				logger.Println(Banner("Skipping blacklist update", err.Error()))
			} else {
				logger.Println("Blacklist updated")
			}
		}

		if settings.FilterInstalled() { // Remove installed packages from process list
			pList.FilterInstalledPkgs(DOC_DIR, GetKernelVersion())
		}

		if settings.RespectGreylist() { // Remove greylisted packages from process list
			pMap.FilterGreylistedPkgs(GREYLIST, &pErr)
		}

		if pList.IsEmpty() {
			stdout.WriteString("Nothing to do\n")
			break
		}

		if !settings.Yes() { // Let user exclude packages and read README files if required
			pList.LetUserExcludePkgs(stdout, os.Stdin, &pErr)
			pList.Iterate(func(pkg *Package) {
				pkg.PromptIfReadmeRequired(stdout, os.Stdin, pager, &pErr)
			})
			if pList.IsEmpty() {
				stdout.WriteString("All packages skipped\n")
				break
			}
		}

		Mkdir(SetEnvIfNotSet("OUTPUT", TMP_DIR)) // Output directory

		executor.PanicOnFail(CLEAN_CMD)

		if FilesToDownload := pList.FilesToDownload(); len(FilesToDownload) > 0 { // Download
			executor.PanicOnFail(DOWNLOAD_CMD + " " + Join(FilesToDownload))
			pErr.CheckDownloads(pList, TMP_DIR) // Check if all files were downloaded
		}

		pList.Iterate(func(pkg *Package) { // Build, and install the packages
			for _, DownloadName := range pkg.DownloadNames() {
				Symlink(
					filepath.Join(TMP_DIR, DownloadName),
					filepath.Join(pkg.Path(), DownloadName),
				)
			}

			// Execute pre-build script if available
			if cmd, err := pkg.PreBuildCmd(CONFIG_DIR); err == nil {
				executor.PanicOnFail(cmd)
			}

			// Build
			if executor.Run(pkg.BuildCmd(settings.BuildOptions()[pkg.Name()])) != nil {
				pErr.SkipPkgAndDependees(pkg, "Failed to build")
			}

			// Install
			executor.PanicOnFail(pkg.InstallCmd(TMP_DIR))
		})

		executor.PanicOnFail(CLEAN_CMD) // Clean up
	}
}
