package main

import (
	"errors"
	"os"
	"testing"
	"time"
)

func TestSetEnvIfNotSet(t *testing.T) {
	if SetEnvIfNotSet("FOO", "bar") != os.Getenv("FOO") {
		t.Error("SetEnvIfNotSet() did not return the correct value")
	}
	if SetEnvIfNotSet("HOME", "bar") != os.Getenv("HOME") || os.Getenv("HOME") == "bar" {
		t.Error("SetEnvIfNotSet() did not return the correct value")
	}
	t.Log("SetEnvIfNotSet() test passed with value:", SetEnvIfNotSet("FOO", "bar"))
}

func TestHandleOSErrorPanic(t *testing.T) {
	defer func() {
		if r := recover(); r == nil {
			t.Error("HandleOSError() did not panic")
		}
	}()
	handleOSError(errors.New("foo"))
}

func TestHandleOSErrorNoPanic(t *testing.T) {
	defer func() {
		if r := recover(); r != nil {
			t.Error("HandleOSError() panicked")
		}
	}()
	handleOSError(nil)
}

func TestCopy(t *testing.T) {
	var (
		td        = t.TempDir()
		expected  = "foo"
		src, dest = Create(td + "/src"), Create(td + "/dest")
		got       = make([]byte, len(expected))
	)
	defer src.Close()
	defer dest.Close()
	src.WriteString(expected)
	src.Seek(0, 0)
	Copy(src, dest)
	dest.Seek(0, 0)
	dest.Read(got)
	if string(got) != expected {
		t.Error("Expected", expected, ", got", got)
	}
}

func TestSymlink(t *testing.T) {
	td := t.TempDir()
	src := td + "/src"
	dest := td + "/dest"
	Create(src).Close()
	Symlink(src, dest)
	if _, err := os.Lstat(dest); err != nil {
		t.Error("Symlink() did not create the symlink")
	}
}

func TestReadDir(t *testing.T) {
	td := t.TempDir()
	files := []string{"bar", "baz", "foo"}
	for _, file := range files {
		Create(td + "/" + file).Close()
	}
	readFiles := ReadDir(td)
	if len(readFiles) != len(files) {
		t.Error("ReadDir() did not return the correct number of files")
	}
	for i, file := range readFiles {
		if file.Name() != files[i] {
			t.Error("Expected", files[i], ", got", file.Name())
		}
	}
}

func TestModTime(t *testing.T) {
	td := t.TempDir()
	f := Create(td + "/foo")
	defer f.Close()
	for _, file := range ReadDir(td) {
		if ModTime(file) == (time.Time{}) {
			t.Error("ModTime() did not return the correct time")
		}
	}
}
