package main

import (
	"bufio"
	"bytes"
	"errors"
	"io"
	"path/filepath"
	"regexp"
	"strings"
	"time"
)

type Package struct {
	name, path string     // path/name = full path
	prev, next *Package   // doubly linked list for processes list
	dependees  []*Package // packages that depend on this package
}

// Getters
func (pkg *Package) Dependees() []*Package  { return pkg.dependees }
func (pkg *Package) Dependencies() []string { return strings.Fields(pkg.queryInfoFile("REQUIRES")) }
func (pkg *Package) InfoFile() string       { return filepath.Join(pkg.Path(), pkg.Name()+".info") }
func (pkg *Package) Name() string           { return pkg.name }
func (pkg *Package) Path() string           { return filepath.Join(pkg.path, pkg.Name()) }
func (pkg *Package) Readme() string         { return filepath.Join(pkg.Path(), "README") }
func (pkg *Package) SlackBuild() string     { return filepath.Join(pkg.Path(), pkg.Name()+".SlackBuild") }
func (pkg *Package) Version() string        { return pkg.queryInfoFile("VERSION") }
func (pkg *Package) FullName() string       { return pkg.Name() + "-" + pkg.Version() }
func (pkg *Package) SlackDesc() string      { return filepath.Join(pkg.Path(), "slack-desc") }

// IsReadmeRequired returns true if the package requires the README file to be read.
func (pkg *Package) IsReadmeRequired() bool {
	matched, err := regexp.MatchString("%README%", pkg.queryInfoFile("REQUIRES"))
	return matched && err == nil
}

// AllReadmes returns a space separted list of all README files in the package
// directory as a string.
func (pkg *Package) AllReadmes() string {
	sb := strings.Builder{}
	re := regexp.MustCompile("README")
	path := pkg.Path()

	for _, file := range ReadDir(path) {
		fn := file.Name()
		if re.MatchString(fn) {
			sb.WriteString(filepath.Join(path, fn) + " ")
		}
	}

	return sb.String()[:sb.Len()-1]
}

func (pkg *Package) DownloadUrls() []string {
	download_x86_64 := pkg.queryInfoFile("DOWNLOAD_x86_64")
	if download_x86_64 != "" {
		return strings.Fields(download_x86_64)
	}
	return strings.Fields(pkg.queryInfoFile("DOWNLOAD"))
}

func (pkg *Package) Md5sums() []string {
	md5sum_x86_64 := pkg.queryInfoFile("MD5SUM_x86_64")
	if md5sum_x86_64 != "" {
		return strings.Fields(md5sum_x86_64)
	}
	return strings.Fields(pkg.queryInfoFile("MD5SUM"))
}

func (pkg *Package) DownloadNames() []string {
	urls := pkg.DownloadUrls()
	names := make([]string, len(urls))
	for i, url := range urls {
		names[i] = filepath.Base(url)
	}
	return names
}

func (pkg *Package) IsUpToDate(docDir, kernelVersion string) bool {
	fullPath := filepath.Join(docDir, pkg.FullName())

	if !Exists(fullPath) {
		if fullPath += "_" + kernelVersion; !Exists(fullPath) {
			return false
		}
	}

	return bytes.Equal(
		ReadFile(pkg.SlackBuild()),
		ReadFile(filepath.Join(fullPath, pkg.Name()+".SlackBuild")),
	)
}

func (pkg *Package) AddDependee(dependee *Package) {
	pkg.dependees = append(pkg.dependees, dependee)
}

func (pkg *Package) queryInfoFile(key string) string {
	file := open(pkg.InfoFile())
	defer file.Close()

	for scanner := bufio.NewScanner(file); scanner.Scan(); {

		if s := strings.Split(scanner.Text(), "\""); s[0] == key+"=" {
			for strings.HasSuffix(s[1], "\\") { // line continuation
				scanner.Scan()
				s[1] = s[1][:len(s[1])-1] + strings.Trim(scanner.Text(), "\"")
			}
			return s[1]
		}

	}

	panic(pkg.Name() + ": " + key + " not found in " + pkg.InfoFile())
}

func (pkg *Package) InstallCmd(tmpDir string) string {
	var (
		re               = regexp.MustCompile(pkg.FullName() + ".*_SBo.*")
		modificationTime time.Time
		pkgName          string
	)
	for _, slackpkg := range ReadDir(tmpDir) {
		if re.MatchString(slackpkg.Name()) {
			if t := ModTime(slackpkg); t.After(modificationTime) {
				modificationTime = t
				pkgName = slackpkg.Name()
			}
		}
	}
	if pkgName == "" {
		panic(pkg.Name() + ": No package to install")
	}
	return INSTALL_CMD + filepath.Join(tmpDir, pkgName)
}

func (pkg *Package) BuildCmd(buildOptions []string) string {
	if len(buildOptions) == 0 {
		return BUILD_CMD + pkg.SlackBuild()
	}
	return ENV_CMD + Join(buildOptions) + " " + BUILD_CMD + pkg.SlackBuild()
}

func (pkg *Package) PreBuildCmd(configDir string) (string, error) {
	filename := filepath.Join(configDir, pkg.Name()+PREBUILD_EXT)
	if IsExecutable(filename) {
		return filename + " " + pkg.Path(), nil
	}
	return "", errors.New(filename + " does not exist or is not executable")
}

// Prompts the user to read the README file if it is listed as required in the
// info file.
func (pkg *Package) PromptIfReadmeRequired(
	w *bufio.Writer,
	r io.Reader,
	pager *Pager,
	PkgErrors *PackageErrors,
) {
	if pkg.IsReadmeRequired() && PromptUser(w, r, pkg.Name()+
		": Requires to read the README file. Do you want "+
		"to read it (and optionally skip the package) now?",
	) {

		pager.Display(pkg.AllReadmes())

		if PromptUser(w, r, "Do you want to skip "+pkg.Name()) {
			PkgErrors.SkipPkgAndDependees(pkg, "User skipped")
		}

	}
}

// Description returns the name plus the description of the package as it is
// found in the first line of the slack-desc file.
func (pkg *Package) Description() string {
	file := open(pkg.SlackDesc())
	defer file.Close()

	prefix := pkg.Name() + ": "
	lenPrefix := len(prefix)

	for scanner := bufio.NewScanner(file); scanner.Scan(); {
		if line := scanner.Text(); strings.HasPrefix(line, prefix) {
			return line[lenPrefix:]
		}
	}

	// We fallback to the name if for some reason the slack-desc file is
	// ill-formatted.
	return pkg.Name()
}

// List related functions
func (pkg *Package) IsInList() bool { return pkg.prev != nil }

// Appends the package to the doubly linked list. Takes the tail of the list as
// an argument and returns the new tail.
func (pkg *Package) Append(tail *Package) *Package {
	tail.next, pkg.prev = pkg, tail
	return pkg // new tail
}

// Delets the package from the doubly linked list. Works also for the head or
// tail of the list.
func (pkg *Package) Del() {
	if pkg.prev != nil {
		pkg.prev.next = pkg.next
	}
	if pkg.next != nil {
		pkg.next.prev = pkg.prev
	}
	pkg.prev = nil
}
