package main

import (
	"bufio"
	"errors"
	"fmt"
	"io"
	"os"
	"path/filepath"
	"strconv"
	"strings"
	"sync"
)

func (tail *Package) ResolvDep(
	parentName string,
	pkgNames []string,
	repoMap RepoMap,
) *Package {
	for _, pkgName := range pkgNames {
		switch pkgName {
		case "%README%", "":
			continue
		case parentName:
			panic("Circular dependency detected: " + parentName)
		}
		pkgPtr := repoMap.Get(pkgName)
		if parentName != "" {
			pkgPtr.AddDependee(repoMap.Get(parentName))
		}
		if pkgPtr.IsInList() {
			continue
		}
		tail = tail.ResolvDep(pkgName, pkgPtr.Dependencies(), repoMap)
		tail = pkgPtr.Append(tail)
	}
	return tail
}

type Head = *Package

func (h Head) Iterate(f func(*Package)) {
	for p := h.next; p != nil; p = p.next {
		f(p)
	}
}

func (h Head) IterateConcurrently(f func(*Package)) {
	var wg sync.WaitGroup
	for p := h.next; p != nil; p = p.next {
		wg.Add(1)
		go func(p *Package) {
			f(p)
			wg.Done()
		}(p)
	}
	wg.Wait()
}

func (h Head) IterateDependees(f func(*Package)) {
	for _, dependee := range h.Dependees() {
		f(dependee)
		dependee.IterateDependees(f)
	}
}

// Removes packages that are already installed. Parallelized.
func (h Head) FilterInstalledPkgs(docDir, kernelVersion string) {
	var mu sync.Mutex
	h.IterateConcurrently(func(pkg *Package) {
		if pkg.IsUpToDate(docDir, kernelVersion) {
			mu.Lock()
			pkg.Del()
			mu.Unlock()
		}
	})
}

// Prompts the user to exclude some packages from the list
func (h Head) LetUserExcludePkgs(w *bufio.Writer, r io.Reader, eh *PackageErrors) {
	// Ask user which packages to exclude
	i := 0
	pkgArray := make([]*Package, 0)
	h.Iterate(func(pkg *Package) {
		pkgArray = append(pkgArray, pkg)
		i++
		fmt.Fprintln(w, i, pkg.Description())
	})
	w.WriteString("Space separated list of packages to exclude:\n")
	w.Flush()
	// Parse user input
	scanner := bufio.NewScanner(r)
	scanner.Scan()
	for _, field := range strings.Fields(scanner.Text()) {
		num, err := strconv.Atoi(field)
		if err != nil || num < 1 || num > i {
			w.WriteString("Invalid input. Repeat...\n")
			h.LetUserExcludePkgs(w, r, eh)
			return
		}
		pkg := pkgArray[num-1]
		pkg.Del()
		eh.Add(pkg, "Excluded by user")
	}
}

func (h Head) UpdateBlacklist(blacklist, otherBlacklist, DefaultFile string) error {
	dest, err := os.Create(blacklist)
	if err != nil {
		return err
	}
	defer dest.Close()

	src, err := os.Open(otherBlacklist)
	if err != nil {
		return errors.New("File " + otherBlacklist + " does not exist: It will be copied to " +
			blacklist + "\nbefore the packages listed in " + DefaultFile +
			" and their dependencies are\nappended to " + blacklist +
			". Either copy your existing " + blacklist + "\nto " + otherBlacklist +
			" or backup " + blacklist + " and create an empty file\n" +
			otherBlacklist + ".")
	}
	defer src.Close()

	Copy(src, dest)
	h.Iterate(func(pkg *Package) { fmt.Fprintln(dest, pkg.Name()) })
	return nil
}

func (h Head) FilesToDownload() []string {
	var (
		FilesToDownload = make([]string, 0)
		mu              sync.Mutex
	)
	h.IterateConcurrently(func(pkg *Package) {
		md5sums := pkg.Md5sums()
		for i, DownloadName := range pkg.DownloadNames() {
			fullPath := filepath.Join(TMP_DIR, DownloadName)
			if !Exists(fullPath) || !IsMD5SumMatch(md5sums[i], fullPath) {
				mu.Lock()
				FilesToDownload = append(FilesToDownload, pkg.DownloadUrls()[i])
				mu.Unlock()
			}
		}
	})
	return FilesToDownload
}

func (h Head) IsEmpty() bool { return h.next == nil || !h.next.IsInList() }
