package main

import (
	"bufio"
	"bytes"
	"os"
	"path/filepath"
	"strings"
	"testing"
)

// --------------------------------------------------
// 1) Test for NO circular dependency
// --------------------------------------------------

func TestResolvDep_NoCircularDependency(t *testing.T) {
	// Suppose 'dependee1' does not actually require itself,
	// so that we won't get a circular dependency.
	dep1 := &Package{name: "dependee1"}
	dep2 := &Package{name: "dependee2"}

	// For dep1, let's require just %README%, no other pkg => no self-reference
	createPkgTestfilesWithRequires(dep1, t, "%README%")

	// For dep2, also no mention of dependee1 => no circular dependency
	createPkgTestfilesWithRequires(dep2, t, "%README%")

	// "parent" is the one from which we start resolving
	parent := &Package{name: "parent"}
	rm := RepoMap{
		"dependee1": dep1,
		"dependee2": dep2,
		"parent":    parent,
	}

	head := &Package{} // sentinel
	tail := head

	// We'll call ResolvDep with pkgNames = ["dependee1", "dependee2"]
	tail = tail.ResolvDep("parent", []string{"dependee1", "dependee2"}, rm)

	// The returned tail should be the last appended package: the chain is head -> dep1 -> dep2
	if head.next == nil || head.next.name != "dependee1" {
		t.Fatalf("Expected head->dependee1, got something else.")
	}
	if head.next.next == nil || head.next.next.name != "dependee2" {
		t.Fatalf("Expected dependee1->dependee2, got something else.")
	}

	// Check that we added parent as a dependee to each
	d1 := rm["dependee1"]
	d2 := rm["dependee2"]
	if len(d1.dependees) == 0 || d1.dependees[0].Name() != "parent" {
		t.Errorf("Expected dependee1 to have parent as dependee, found: %+v", d1.dependees)
	}
	if len(d2.dependees) == 0 || d2.dependees[0].Name() != "parent" {
		t.Errorf("Expected dependee2 to have parent as dependee, found: %+v", d2.dependees)
	}
}

// --------------------------------------------------
// 2) Test for CIRCULAR dependency
// --------------------------------------------------

func TestResolvDep_CircularDependency(t *testing.T) {
	// For circular dependency, let 'dependee1' require itself:
	dep1 := &Package{name: "dependee1"}
	dep2 := &Package{name: "dependee2"}

	// E.g., REQUIRES="%README% dependee1"
	// That means 'dependee1' is said to require 'dependee1' again => circular
	createPkgTestfilesWithRequires(dep1, t, "%README% dependee1")
	createPkgTestfilesWithRequires(dep2, t, "%README%")

	rm := RepoMap{
		"dependee1": dep1,
		"dependee2": dep2,
		"parent":    &Package{name: "parent"},
	}

	head := &Package{}
	tail := head

	// We expect a panic due to circular dependency
	defer func() {
		if r := recover(); r == nil {
			t.Errorf("Expected panic due to circular dependency, but no panic occurred.")
		} else {
			// Optionally check the message
			msg, ok := r.(string)
			if ok && !containsAll(msg, "Circular dependency", "dependee1") {
				t.Errorf("Expected panic to mention 'Circular dependency' and 'dependee1', got %q", msg)
			}
		}
	}()

	tail = tail.ResolvDep("parent", []string{"dependee1", "dependee2"}, rm)
	t.Errorf("Should have panicked before reaching this line.")
}

func TestResolvDep_AlreadyInList(t *testing.T) {
	// 1) Create the "dependee1" package whose .info does NOT create a circular dependency.
	//    For simplicity, we only have: REQUIRES="%README%" (no mention of itself).
	dep1 := &Package{name: "dependee1"}
	createPkgTestfilesWithRequires(dep1, t, "%README%")

	// 2) Create a "parent" package in the RepoMap.
	parent := &Package{name: "parent"}
	rm := RepoMap{
		"dependee1": dep1,
		"parent":    parent,
	}

	// 3) Build an initial chain: head -> dep1 (so dep1 is already "in the list").
	head := &Package{} // sentinel
	tail := head
	tail = dep1.Append(tail) // Now head->dep1

	// 4) Now call ResolvDep with "dependee1" again. This should trigger:
	//    if pkgPtr.IsInList() { continue }
	//    but also ensure pkgPtr.AddDependee(parent) still happens.
	tail = tail.ResolvDep("parent", []string{"dependee1"}, rm)

	// 5) Verify "dependee1" has "parent" as a dependee
	if len(dep1.dependees) == 0 || dep1.dependees[0].Name() != "parent" {
		t.Errorf("Expected 'dependee1' to have 'parent' as a dependee, got %+v",
			dep1.dependees)
	}

	// 6) Verify the chain is still just head->dep1 (i.e., 'dependee1' was NOT appended again).
	if head.next != dep1 || dep1.next != nil {
		t.Errorf("Expected chain to remain [head->dependee1], but got head->%v->%v",
			head.next, dep1.next)
	}
}

func TestIterate(t *testing.T) {
	p := mockPkg()
	// mockPkg returns [prev -> name -> next], with p = "name"
	var names []string
	p.prev.Iterate(func(pkg *Package) {
		names = append(names, pkg.Name())
	})
	// We expect "name" then "next" in the iteration (since iteration starts at head.next)
	if len(names) != 2 || names[0] != "name" || names[1] != "next" {
		t.Errorf("Expected iteration order [name, next], got %v", names)
	}
}

func TestIterateDependees(t *testing.T) {
	p := mockPkg()
	// In your code, p has dependees = [prev, next]
	var collected []string
	p.IterateDependees(func(d *Package) {
		collected = append(collected, d.Name())
	})
	// The default mock has two dependees: "prev" and "next".
	// There's no recursion in their dependees by default, but if you want,
	// you can add some nested dependees to test deeper recursion.
	want := []string{"prev", "next"}
	if !equalStringSlices(collected, want) {
		t.Errorf("Expected dependees=[%v], got %v", want, collected)
	}
}

func TestFilterInstalledPkgs(t *testing.T) {
	// Step 1: Create the minimal chain [prev -> name -> next]
	p := mockPkg()
	prev, next := p.prev, p.next
	head := p.prev // (head is actually "prev" itself; it’s the sentinel).
	// By default, mockPkg returns p = "name" with p.prev = "prev", p.next = "next"
	// Create .info, .SlackBuild, etc. for each package, so they have real files for IsUpToDate checks.
	createPkgTestfiles(p, t)    // for "name"
	createPkgTestfiles(prev, t) // for "prev"
	createPkgTestfiles(next, t) // for "next"

	// Step 2: Initially, none of them are "up-to-date", so filtering should remove nothing.
	head.FilterInstalledPkgs("/doc/dir", "kernel-1.0")

	// Ensure the chain is still [prev -> name -> next]
	if !prev.next.IsInList() || prev.next.name != "name" {
		t.Errorf("Expected 'name' to remain in the chain, but it's missing or not in list.")
	}
	if !prev.next.next.IsInList() || prev.next.next.name != "next" {
		t.Errorf("Expected 'next' to remain in the chain, but it's missing or not in list.")
	}

	// Step 3: Now simulate "prev" being up-to-date. According to IsUpToDate:
	//         docDir + pkg.FullName() OR docDir + pkg.FullName() + "_" + kernelVersion
	//         must exist, and SlackBuild must match (MD5) with the one in the docDir.
	docDir := t.TempDir()

	// We'll assume "next" has version "1.0" from the .info content. By default, FullName = "next-1.0".
	// Let's create docDir/next-1.0 and copy the SlackBuild file there.
	fullPath := filepath.Join(docDir, next.FullName()) // typically "next-1.0"
	mustMkdir(t, fullPath)

	// Copy the SlackBuild from the package to that location
	srcSlackBuild := open(next.SlackBuild()) // e.g., /tmp/someRandomDir/next.SlackBuild
	defer srcSlackBuild.Close()
	dstSlackBuildPath := filepath.Join(fullPath, next.Name()+".SlackBuild")
	dstSlackBuild := Create(dstSlackBuildPath)
	defer dstSlackBuild.Close()
	Copy(srcSlackBuild, dstSlackBuild) // so MD5 sums match

	// Step 4: Filter again, now "next" should be removed from the list (IsUpToDate returns true).
	head.FilterInstalledPkgs(docDir, "kernel-1.0")

	// The chain should now skip 'next', leaving [prev -> name]
	if next.IsInList() {
		t.Errorf("Expected 'next' to be removed from the list, but it's still considered in list.")
	}
	// Meanwhile, 'name' and 'prev' should still be in the list
	if head.next == nil || head.next.name != "name" || !head.next.IsInList() {
		t.Errorf("Expected chain to have 'name' after head, but got something else.")
	}
	if p.prev == nil || p.prev.name != "prev" || !p.prev.IsInList() {
		t.Errorf("Expected chain to have 'prev' before 'name', but got something else.")
	}
}

func TestLetUserExcludePkgs(t *testing.T) {
	p := mockPkg()
	// This chain is [prev -> name -> next], with head=prev
	head := p.prev
	createPkgTestfiles(p, t)
	createPkgTestfiles(p.next, t)

	var buf bytes.Buffer
	writer := bufio.NewWriter(&buf)
	reader := strings.NewReader("2\n") // user input: "2" => exclude the 2nd package in the listing

	var errs PackageErrors

	// The function prints a list of packages (1, 2, 3, ...) then reads user input
	// In our chain, iteration sees next packages: [name, next]
	// So "1" => name, "2" => next
	head.LetUserExcludePkgs(writer, reader, &errs)
	writer.Flush()

	if len(errs) != 1 {
		t.Fatalf("Expected 1 error entry, got %d", len(errs))
	}
	if !strings.Contains(errs[0], "next") || !strings.Contains(errs[0], "Excluded by user") {
		t.Errorf("Expected 'next: Excluded by user', got %q", errs[0])
	}

	// Ensure "next" was removed from chain, only "name" remains after head
	if head.next == nil || head.next.name != "name" {
		t.Errorf("Expected chain to be head->name, got something else")
	}
	if head.next.next != nil {
		t.Errorf("Expected end of chain after 'name', but found %v", head.next.next)
	}
}

func TestLetUserExcludePkgs_InvalidInput(t *testing.T) {
	p := mockPkg()
	head := p.prev
	createPkgTestfiles(p, t)
	createPkgTestfiles(p.next, t)

	// We'll input "99", which is invalid (out of range).
	var output bytes.Buffer
	writer := bufio.NewWriter(&output)
	reader := strings.NewReader("99\n") // invalid input

	var errs PackageErrors
	head.LetUserExcludePkgs(writer, reader, &errs)
	writer.Flush()

	// Because it's invalid, the function tries again recursively.
	// The easiest check is to confirm that "Invalid input. Repeat..." is in the output
	outStr := output.String()
	if !strings.Contains(outStr, "Invalid input. Repeat...") {
		t.Errorf("Expected 'Invalid input. Repeat...' in output, got:\n%s", outStr)
	}
	// We won't re-check final chain state or errs because the function restarts.
}

func TestUpdateBlacklist(t *testing.T) {
	// We'll test that UpdateBlacklist copies otherBlacklist -> blacklist
	// and then appends the packages from the chain to the end.

	// Prepare a chain: [p1, p2]
	head := &Package{}
	p1 := &Package{name: "p1"}
	p2 := &Package{name: "p2"}
	p1.Append(head)
	p2.Append(p1)

	// Create two temp files: one for 'otherBlacklist' (source), one for 'blacklist' (destination)
	tdir := t.TempDir()
	blacklistPath := filepath.Join(tdir, "blacklist.txt")
	otherPath := filepath.Join(tdir, "other.txt")
	defaultFile := "someDefaultFile" // not strictly used in success path, but part of error text

	// Write some lines to other.txt
	if err := os.WriteFile(otherPath, []byte("alreadyThere1\nalreadyThere2\n"), 0o644); err != nil {
		t.Fatalf("failed to write other.txt: %v", err)
	}

	// Now call UpdateBlacklist
	err := head.UpdateBlacklist(blacklistPath, otherPath, defaultFile)
	if err != nil {
		t.Fatalf("unexpected error: %v", err)
	}

	// Check that blacklist.txt now contains original lines plus p1, p2
	content, err := os.ReadFile(blacklistPath)
	if err != nil {
		t.Fatalf("failed to read blacklist.txt: %v", err)
	}
	lines := strings.Split(strings.TrimSpace(string(content)), "\n")
	wantLines := []string{"alreadyThere1", "alreadyThere2", "p1", "p2"}
	if !equalStringSlices(lines, wantLines) {
		t.Errorf("expected lines %v in blacklist, got %v", wantLines, lines)
	}
}

func TestUpdateBlacklist_MissingOtherFile(t *testing.T) {
	// If otherBlacklist doesn’t exist, we expect an error with certain text
	head := &Package{}
	tdir := t.TempDir()
	blacklistPath := filepath.Join(tdir, "blacklist.txt")
	otherPath := filepath.Join(tdir, "doesnotexist.txt")

	err := head.UpdateBlacklist(blacklistPath, otherPath, "defaultFile")
	if err == nil {
		t.Fatalf("expected an error for missing otherBlacklist, got nil")
	}
	if !strings.Contains(err.Error(), "does not exist") {
		t.Errorf("expected 'does not exist' in error, got %v", err)
	}
}

func TestUpdateBlacklist_PermissionDenied(t *testing.T) {
	// 1) Create a root temp directory.
	rootDir := t.TempDir()

	// 2) Create a *subdirectory* in which we *can* write files
	//    for the "otherBlacklist" file.
	otherDir := filepath.Join(rootDir, "other_dir")
	if err := os.Mkdir(otherDir, 0o755); err != nil {
		t.Fatalf("failed to create otherDir: %v", err)
	}

	// 3) Create another subdirectory for the "blacklist" file,
	//    but remove write permissions from it so `os.Create` fails inside it.
	noWriteDir := filepath.Join(rootDir, "no_write_dir")
	if err := os.Mkdir(noWriteDir, 0o755); err != nil {
		t.Fatalf("failed to create noWriteDir: %v", err)
	}

	// 4) Create the other file *before* removing write permission from noWriteDir
	otherFilePath := filepath.Join(otherDir, "other.txt")
	if err := os.WriteFile(otherFilePath, []byte("existing line\n"), 0o644); err != nil {
		t.Fatalf("failed to create other.txt: %v", err)
	}

	// 5) Now remove write permissions from noWriteDir.
	if err := os.Chmod(noWriteDir, 0o500); err != nil {
		t.Fatalf("failed to chmod the noWriteDir: %v", err)
	}

	// 6) Our “blacklist” path will be inside noWriteDir, which is not writable.
	blacklistPath := filepath.Join(noWriteDir, "blacklist.txt")

	// 7) Prepare a Head package
	head := &Package{name: "headPkg"}

	// 8) Call UpdateBlacklist - expect "permission denied" when trying to create blacklist.txt
	err := head.UpdateBlacklist(blacklistPath, otherFilePath, "defaultFile")
	if err == nil {
		t.Fatalf("expected an error due to insufficient permissions, got nil")
	}

	// 9) Optionally check the error text
	t.Logf("Got error: %v", err)
	if !strings.Contains(strings.ToLower(err.Error()), "permission") &&
		!strings.Contains(strings.ToLower(err.Error()), "denied") {
		t.Errorf("expected error mentioning 'permission denied', got: %v", err)
	}
}

func TestFilesToDownload(t *testing.T) {
	// Let’s create a chain [p1, p2]
	head := &Package{}
	p1 := &Package{name: "p1"}
	p2 := &Package{name: "p2"}
	createPkgTestfiles(p1, t) // create some files for p1
	createPkgTestfiles(p2, t) // create some files for p2
	p1.Append(head)
	p2.Append(p1)

	urls := head.FilesToDownload()
	// Each package has 1 DownloadNames => total 2 URLs to download
	if len(urls) != 2 {
		t.Fatalf("expected 4 URLs, got %d", len(urls))
	}
	// Each URL is "http://example.com/foo.tar.gz" by default from our stub
	for i, u := range urls {
		if u != "http://example.com/foo.tar.gz" {
			t.Errorf("expected 'http://example.com/foo.tar.gz', got %q at index %d", u, i)
		}
	}
}

func TestIsEmpty(t *testing.T) {
	head := &Package{}
	if !head.IsEmpty() {
		t.Errorf("Expected empty head to be IsEmpty=true, got false")
	}

	// Once we append a package, it should not be empty
	p := &Package{name: "pkg1"}
	p.Append(head)
	if head.IsEmpty() {
		t.Errorf("Expected head to not be empty after append, got IsEmpty=true")
	}

	// If we remove the package, it should be empty again
	p.Del()
	if !head.IsEmpty() {
		t.Errorf("Expected head to be empty after deletion, got IsEmpty=false")
	}
}
