package main

import (
	"io/fs"
	"path/filepath"
	"sync"
)

type RepoMap map[string]*Package

// Returns a map containing infofile[package_name] = package
// Since this routine is usually the bottleneck, it is parallelized.
func NewRepoMap(RepoPath string) RepoMap {

	var (
		wg   sync.WaitGroup
		mu   sync.Mutex
		repo = make(RepoMap, 10100)
	)

	for _, file := range ReadDir(RepoPath) {
		// We parallelize over the `categories` in the repository
		wg.Add(1)
		go func(file fs.DirEntry) {
			defer wg.Done()

			if !file.IsDir() {
				return
			}

			subdirPath := filepath.Join(RepoPath, file.Name())
			subfiles := ReadDir(subdirPath)

			// We fill a local slice that we merge into the actual map
			// so that we only need to lock the map once.
			local_repo := make([]Package, 0, len(subfiles))
			for _, subfile := range subfiles {
				pkgName := subfile.Name()
				local_repo = append(local_repo, Package{name: pkgName, path: subdirPath})
			}

			mu.Lock() // Merge
			for i := range local_repo {
				p := &local_repo[i]
				repo[p.name] = p
			}
			mu.Unlock()
		}(file)
	}

	wg.Wait()
	return repo
}

func (rm RepoMap) Get(pkgname string) *Package {
	if pkg, ok := rm[pkgname]; ok {
		return pkg
	}
	panic(Banner("Fatel Error", "Package not found in the repository: "+pkgname))
}

func (rm RepoMap) Contains(pkgname string) bool { _, ok := rm[pkgname]; return ok }

// Wrapper around ResolvDep for the uppermost level of recursion
func (rm RepoMap) ProcessList(pkgNames []string) Head {
	pkgs := &Package{}
	pkgs.ResolvDep("", pkgNames, rm)
	return pkgs
}

// Returns a list of packages from a list of package names without resolving dependencies
func (rm RepoMap) PkgNames2List(pkgNames []string) Head {
	head := &Package{}
	tail := head
	for _, pkgName := range pkgNames {
		tail = rm.Get(pkgName).Append(tail)
	}
	return head
}

// Removes packages listed in a file with the given path `greylist`
func (rm RepoMap) FilterGreylistedPkgs(greylist string, PkgErrors *PackageErrors) {
	for _, greylisted := range ReadPkgNamesFromFile(greylist) {
		if pkg := rm.Get(greylisted); pkg.IsInList() {
			pkg.Del()
			PkgErrors.Add(pkg, "Greylisted")
		}
	}
}
