package main

import "flag"

type command int

const ( // Command types, only one of these can be used at a time
	DEPENDEES command = iota
	LIST_INSTALLED
	SEARCH
	TRACKING
	UPGRADE
	VIEW
	VIEW_BUILD
)

type settings struct {
	update, yes, updateBlacklist, checkMD5Sum, respectGreylist, filterInstalled, noDeps bool
	root                                                                                string
	pkgNames                                                                            []string
	buildOptions                                                                        map[string][]string
	cmd                                                                                 command
}

func NewSettings(args []string, DefaultPkgFile string) *settings {
	var (
		fs                 = flag.NewFlagSet(args[0], flag.ExitOnError)
		filterInstalledPtr = fs.Bool(FILTERINSTALLED_FLAG, false, "don't filter out installed packages")
		filePtr            = fs.String(FILE_FLAG, DefaultPkgFile, "path to the input file")
		listInstalledPtr   = fs.Bool(LIST_INSTALLED_FLAG, false, "list installed packages")
		md5Ptr             = fs.Bool(MD5_FLAG, false, "skip MD5Sum checks")
		noDepsPtr          = fs.Bool(NO_DEPS_FLAG, false, "don't install dependencies")
		rootPtr            = fs.String(ROOT_FLAG, "", "display dependees of the given package")
		skipGreylistPtr    = fs.Bool(RESPECTGREYLIST_FLAG, false, "ignore the greylist")
		searchPtr          = fs.Bool(SEARCH_FLAG, false, "search for a package")
		trackPtr           = fs.Bool(TRACK_FLAG, false, "track the given packages (show all dependencies)")
		updatePtr          = fs.Bool(UPDATE_FLAG, false, "update the repository")
		viewPtr            = fs.Bool(VIEW_FLAG, false, "display information for the given packages")
		viewBuildPtr       = fs.Bool(VIEWBUILD_FLAG, false, "display build scripts for the given packages")
		yesPtr             = fs.Bool(YES_FLAG, false, "Skip all prompts")
		IsCmdSet           = false
	)

	fs.Parse(args[1:])

	for _, cmd_flag := range []*bool{viewPtr, viewBuildPtr, trackPtr, searchPtr, listInstalledPtr} {
		if *cmd_flag && IsCmdSet {
			panic("At most one of " + VIEW_FLAG + ", " + VIEWBUILD_FLAG + ", " + TRACK_FLAG +
				", " + LIST_INSTALLED_FLAG + " and " + SEARCH_FLAG + " can be used at a time")
		}
		if !IsCmdSet && *cmd_flag {
			IsCmdSet = true
		}
	}

	pkgNames, buildOptions := PkgNamesAndBuildOptions(fs.Args())
	pkgs_from_file := len(pkgNames) == 0
	if !pkgs_from_file && *filePtr != DefaultPkgFile {
		panic("Cannot use positional arguments and -f together")
	}

	var cmd command
	switch {
	case *listInstalledPtr:
		cmd = LIST_INSTALLED
	case *rootPtr != "":
		cmd = DEPENDEES
	case *searchPtr:
		cmd = SEARCH
	case *trackPtr:
		cmd = TRACKING
	case *viewPtr:
		cmd = VIEW
	case *viewBuildPtr:
		cmd = VIEW_BUILD
	default:
		cmd = UPGRADE
	}

	if pkgs_from_file { // Read the list of packages to upgrade
		pkgNames, buildOptions = PkgNamesAndBuildOptions(ReadPkgNamesFromFile(*filePtr))
	}

	return &settings{
		buildOptions:    buildOptions,
		cmd:             cmd,
		checkMD5Sum:     !*md5Ptr,
		noDeps:          *noDepsPtr,
		filterInstalled: !*filterInstalledPtr,
		respectGreylist: !*skipGreylistPtr,
		root:            *rootPtr,
		pkgNames:        pkgNames,
		updateBlacklist: (*filePtr == DefaultPkgFile) && pkgs_from_file && !*noDepsPtr,
		update:          *updatePtr,
		yes:             *yesPtr,
	}
}

// Getters
func (s *settings) BuildOptions() map[string][]string { return s.buildOptions }
func (s *settings) Cmd() command                      { return s.cmd }
func (s *settings) CheckMD5Sum() bool                 { return s.checkMD5Sum }
func (s *settings) NoDeps() bool                      { return s.noDeps }
func (s *settings) FilterInstalled() bool             { return s.filterInstalled }
func (s *settings) RespectGreylist() bool             { return s.respectGreylist }
func (s *settings) Root() string                      { return s.root }
func (s *settings) PkgNames() []string                { return s.pkgNames }
func (s *settings) UpdateBlacklist() bool             { return s.updateBlacklist }
func (s *settings) Update() bool                      { return s.update }
func (s *settings) Yes() bool                         { return s.yes }
