package main

import "testing"

func TestNewSettings_Commands(t *testing.T) {
	// We test that each mutually exclusive command sets the correct .cmd field

	defaultFile := createTestFile(t.TempDir()+"/pkglist", "pkg1\npkg2\n")

	t.Run("LIST_INSTALLED", func(t *testing.T) {
		args := []string{"test", "-" + LIST_INSTALLED_FLAG}
		s := NewSettings(args, defaultFile)
		if s.Cmd() != LIST_INSTALLED {
			t.Errorf("Expected cmd to be LIST_INSTALLED, got %v", s.Cmd())
		}
	})

	t.Run("DEPENDEES", func(t *testing.T) {
		args := []string{"test", "-" + ROOT_FLAG, "rootPackage"}
		s := NewSettings(args, defaultFile)
		if s.Cmd() != DEPENDEES {
			t.Errorf("Expected cmd to be DEPENDEES, got %v", s.Cmd())
		}
		if s.Root() != "rootPackage" {
			t.Errorf("Expected rootPackage as .root, got %v", s.Root())
		}
	})

	t.Run("SEARCH", func(t *testing.T) {
		args := []string{"test", "-" + SEARCH_FLAG}
		s := NewSettings(args, defaultFile)
		if s.Cmd() != SEARCH {
			t.Errorf("Expected cmd to be SEARCH, got %v", s.Cmd())
		}
	})

	t.Run("TRACKING", func(t *testing.T) {
		args := []string{"test", "-" + TRACK_FLAG}
		s := NewSettings(args, defaultFile)
		if s.Cmd() != TRACKING {
			t.Errorf("Expected cmd to be TRACKING, got %v", s.Cmd())
		}
	})

	t.Run("VIEW", func(t *testing.T) {
		args := []string{"test", "-" + VIEW_FLAG}
		s := NewSettings(args, defaultFile)
		if s.Cmd() != VIEW {
			t.Errorf("Expected cmd to be VIEW, got %v", s.Cmd())
		}
	})

	t.Run("VIEW_BUILD", func(t *testing.T) {
		args := []string{"test", "-" + VIEWBUILD_FLAG}
		s := NewSettings(args, defaultFile)
		if s.Cmd() != VIEW_BUILD {
			t.Errorf("Expected cmd to be VIEW_BUILD, got %v", s.Cmd())
		}
	})

	t.Run("DEFAULT_UPGRADE", func(t *testing.T) {
		// If none of the “command” flags are set and root is empty => UPGRADE
		args := []string{"test"}
		s := NewSettings(args, defaultFile)
		if s.Cmd() != UPGRADE {
			t.Errorf("Expected cmd to be UPGRADE, got %v", s.Cmd())
		}
	})
}

func TestNewSettings_MutuallyExclusiveCommands(t *testing.T) {
	// Try passing both -v and -search => expect panic
	defaultFile := createTestFile(t.TempDir()+"/pkglist", "pkg1\npkg2\n")
	defer func() {
		if r := recover(); r == nil {
			t.Errorf("Expected panic if multiple command flags are used, but got none")
		}
	}()
	args := []string{"test", "-" + VIEW_FLAG, "-" + SEARCH_FLAG}
	_ = NewSettings(args, defaultFile)
	t.Errorf("Should have panicked before this line") // For clarity
}

func TestNewSettings_FileFlagAndPositionalArgs(t *testing.T) {
	defaultFile := createTestFile(t.TempDir()+"/pkglist", "pkg1\npkg2\n")
	t.Run("PanicIfFileFlagAndPositionalArgsUsedTogether", func(t *testing.T) {
		// If we pass -f something AND positional arguments are also given,
		// we expect a panic if the filePtr != DefaultPkgFile
		defer func() {
			if r := recover(); r == nil {
				t.Errorf("Expected panic if both -f customFile and positional arguments used, but got none")
			}
		}()
		args := []string{"test", "-" + FILE_FLAG, "someOtherFile", "pkg1", "pkg2"}
		_ = NewSettings(args, defaultFile)
		t.Errorf("Should have panicked before this line")
	})

	t.Run("NoPanicIfUsingDefaultFileWithoutPositionalArgs", func(t *testing.T) {
		// This should be fine: no positional args => we load from defaultFile
		args := []string{"test", "-" + FILE_FLAG, defaultFile}
		s := NewSettings(args, defaultFile)
		if len(s.PkgNames()) == 2 {
			// By default, if pkgNames is empty, we read from file
		} else {
			t.Errorf("Expected no pkgNames if none are specified, got %v", s.PkgNames())
		}
	})
}

func TestNewSettings_BooleanFlags(t *testing.T) {
	defaultFile := createTestFile(t.TempDir()+"/pkglist", "pkg1\npkg2\n")
	t.Run("MD5Flag", func(t *testing.T) {
		// By default, checkMD5Sum = true, but if we pass -md5 then it's toggled to false
		args := []string{"test", "-" + MD5_FLAG}
		s := NewSettings(args, defaultFile)
		if s.CheckMD5Sum() {
			t.Errorf("Expected checkMD5Sum to be false, got true (md5 flag was set)")
		}
	})

	t.Run("NoDeps", func(t *testing.T) {
		args := []string{"test", "-" + NO_DEPS_FLAG}
		s := NewSettings(args, defaultFile)
		if !s.NoDeps() {
			t.Errorf("Expected noDeps to be true, got false")
		}
	})

	t.Run("FilterInstalled", func(t *testing.T) {
		// The code has filterInstalled as !*filterInstalledPtr
		// meaning the default is true, but if we pass the flag, we flip it.
		args := []string{"test", "-" + FILTERINSTALLED_FLAG}
		s := NewSettings(args, defaultFile)
		if s.FilterInstalled() {
			t.Errorf("Expected filterInstalled to be false, got true")
		}
	})

	t.Run("RespectGreylist", func(t *testing.T) {
		// The code sets s.respectGreylist = !*skipGreylistPtr
		// If we pass -respect-greylist (skipGreylistPtr = true), then respectGreylist = false
		args := []string{"test", "-" + RESPECTGREYLIST_FLAG}
		s := NewSettings(args, defaultFile)
		if s.RespectGreylist() {
			t.Errorf("Expected respectGreylist=false, got true")
		}
	})

	t.Run("Update", func(t *testing.T) {
		args := []string{"test", "-" + UPDATE_FLAG}
		s := NewSettings(args, defaultFile)
		if !s.Update() {
			t.Errorf("Expected update=true, got false")
		}
	})

	t.Run("Yes", func(t *testing.T) {
		args := []string{"test", "-" + YES_FLAG}
		s := NewSettings(args, defaultFile)
		if !s.Yes() {
			t.Errorf("Expected yes=true, got false")
		}
	})
}

func TestNewSettings_UpdateBlacklist(t *testing.T) {
	// s.updateBlacklist gets set if:
	//   (filePtr == DefaultPkgFile) && pkgs_from_file && !noDepsPtr
	// That means:
	//   1) -f not used or is defaultFile
	//   2) pkgNames are empty => loaded from file => pkgs_from_file = true
	//   3) -no-deps not passed

	defaultFile := createTestFile(t.TempDir()+"/pkglist", "pkg1\npkg2\n")

	t.Run("UpdateBlacklist_True", func(t *testing.T) {
		args := []string{"test"} // no commands => upgrade, no file override
		// No positional arguments => pkgs_from_file is true
		// filePtr == DefaultPkgFile => let's say defaultFile
		s := NewSettings(args, defaultFile)
		if !s.UpdateBlacklist() {
			t.Errorf("Expected updateBlacklist=true under these conditions, got false")
		}
	})

	t.Run("UpdateBlacklist_False_When_CustomFile", func(t *testing.T) {
		// If we use -f customFile, then (filePtr == DefaultPkgFile) is false
		// => s.updateBlacklist should be false
		customFile := createTestFile(t.TempDir()+"/pkglist", "pkg1\npkg2\n")
		args := []string{"test", "-" + FILE_FLAG, customFile}
		s := NewSettings(args, defaultFile)
		if s.UpdateBlacklist() {
			t.Errorf("Expected updateBlacklist=false when using -f customFile, got true")
		}
	})

	t.Run("UpdateBlacklist_False_When_PositionalArgs", func(t *testing.T) {
		// If we pass positional arguments => pkgs_from_file = false
		// => s.updateBlacklist = false
		args := []string{"test", "pkg1", "pkg2"}
		s := NewSettings(args, defaultFile)
		if s.UpdateBlacklist() {
			t.Errorf("Expected updateBlacklist=false when positional arguments are provided, got true")
		}
	})

	t.Run("UpdateBlacklist_False_When_NoDeps", func(t *testing.T) {
		args := []string{"test", "-" + NO_DEPS_FLAG}
		s := NewSettings(args, defaultFile)
		if s.UpdateBlacklist() {
			t.Errorf("Expected updateBlacklist=false when noDeps is set, got true")
		}
	})
}

func TestNewSettings_PkgNamesAndBuildOptions(t *testing.T) {
	// Suppose we pass multiple packages with build options

	defaultFile := createTestFile(t.TempDir()+"/pkglist", "pkg1\npkg2\n")

	args := []string{"test", "pkg1", "OPT1=val1", "pkg2", "OPT2=val2"}
	s := NewSettings(args, defaultFile)
	pkgs := s.PkgNames()
	if len(pkgs) != 2 {
		t.Fatalf("Expected 2 packages, got %d", len(pkgs))
	}
	if pkgs[0] != "pkg1" || pkgs[1] != "pkg2" {
		t.Errorf("Expected pkg1, pkg2; got %v", pkgs)
	}
	if got := s.BuildOptions()["pkg1"]; !compareStringSlices(got, []string{"OPT1=val1"}) {
		t.Errorf("Expected buildOptions for pkg1=[OPT1=val1], got %v", got)
	}
	if got := s.BuildOptions()["pkg2"]; !compareStringSlices(got, []string{"OPT2=val2"}) {
		t.Errorf("Expected buildOptions for pkg2=[OPT2=val2], got %v", got)
	}
}

// ----------------------------------------------------------------
// Helpers
// ----------------------------------------------------------------
func compareStringSlices(a, b []string) bool {
	if len(a) != len(b) {
		return false
	}
	for i := range a {
		if a[i] != b[i] {
			return false
		}
	}
	return true
}

func createTestFile(fname, content string) string {
	f := Create(fname)
	defer f.Close()
	f.WriteString(content)
	return fname
}
